function [A,X,A_log,offdiag,it] = JTJD(C,tol)

% INPUT   C : the set of n  m x m matrices (in a 3D array) to be jointly diagonalized, i.e., target
%             matrices  
%         tol : the convergence criterion, which is 1e-6 by default
% OUTPUT  A : the estimate of the joint diagonalizer
%         X : the diagonalized matrix set
%         A_log : the record list of A obtained after each sweep
%         offdiag : the record list of the norms of the off-diagonal part
%                  of the target matrices per sweep
%         it : the total number of sweeps before convergence
%
% X. Guo, S. Zhu, S. Miron and D. Brie, "Approximate Joint Diagonalization
% by Nonorthogonal Nonparametric Jacobi Transformations", to appear in
% ICASSP 2010, Dallas, TX.
% 
% The Matlab code is implemented by X. Guo. 
% Please contact: xijing.guo@cran.uhp-nancy.fr 
% if you have any questions or suggestions.
%
%
% Here we directly use the cost_off function from the ffdiag_pack
% implemented by Dr.Ziehe et al.

[m m n] = size(C);
X = reshape(C,[m m*n]);

it	= 1;
Maxit = 20;
if nargin <2, tol = 1e-6; end
A = eye(m); 
A_log = cell(1,Maxit);
offdiag = zeros(1,Maxit);

while it<Maxit
    % Jacobi transformations
    for p=1:m-1,
        for q=p+1:m,

            pp = p:m:n*m ;
            qq = q:m:n*m ;           
            x = [ X(p,pp) ; X(q,pp) ; X(p,qq) ; X(q,qq)] ; 
            [v,d] = eig(x*x');
            [dd, k] = sort(diag(d));
            
            P = reshape(v(:,k(3)),[2 2]); 
            Q = reshape(v(:,k(4)),[2 2]);
            t = det(Q) ;

            % Compute G = P/Q
            a = det([P(2,:);Q(2,:)]);
            b1 = det([P(1,:);Q(2,:)]);
            b2 = det([Q(1,:);P(2,:)]);           
            c = det([Q(1,:);P(1,:)]);
            G = [b1 c; a b2]/t;
            
            
% The following lines (commented) are the implementation of the "closed 
% form" solution suggested in the paper to form the elementary
% transformation matrix. Nontheless, we encounter an numerical stability
% problem with this design, though it is theorectically the same as we use
% the Matlab standard funtion eig(); see Matlab Help on "eig".
%
% The main problem happens when we use the following code for the
% eigenvectos of G is with the normalization of the 2 by 2 eigenmatrix,
% which is necessary to avoid trivial solutions. The row-wise normalization
% can be problematic if the norms of the rows are too small, beyond the
% precision limit, then error occurs, as in this case the normalization
% process undergoes subdivsion by a factor that is very close to zero. In
% contrast, the standard Matlab function eig(), is designed
% to retrive normailzed solution without such problems.  
%
% At present, we are investigating the possible solutions to handle this
% normalization problem. Yet, before we could reach any solution, we will
% use eig() for its computational rubostness. 
%
% We are in debt to the interested readers who are willing to help with
% their valuable suggestions, and here by thank them in advance. 

%             b = b1 - b2;
%             if isreal(t)
%                 s = sign(t);
%             end
%             f = sqrt(b*b+4*a*c);
% %             u = [2*c s*f-b;s*f-b -2*a]; 
% %             v = inv(normit(u));
%             v = [-2*a -s*f+b; -s*f+b 2*c];
%             v = normit(v')';

            [u,dd] = eig(G);
            v = inv(u);
            
            % Update A and the target matrices X
            h = [p,q];
            k = [pp,qq];

            A(h,:) 	= v * A(h,:) ;
            
            
            X(h,:)	= v * X(h,:) ;
            u = v';
            X(:,k) = [ u(1)*X(:,pp)+u(2)*X(:,qq) u(3)*X(:,pp)+u(4)*X(:,qq) ] ;
        end
    end
    
    g = cost_off(C,normit(A')');
    
    offdiag(it) = g; 
    A_log{it} = A;  
    
%     fprintf(1,'iteriation %d :  %6.8f  \n ',it, g);
    
    if it>1 && abs(g-offdiag(it-1)) < tol
        break;
    end 
    it = it + 1;
end
